// clang-format off
#ifdef __CINT__

#pragma link off all globals;
#pragma link off all classes;
#pragma link off all functions;

#pragma link C++ class MFileHist+;
#pragma link C++ class VMatrix+;
#pragma link C++ class MFMatrix+;
#pragma link C++ class RMatrix+;
#pragma link C++ class MatOp+;

#endif
