// clang-format off
#ifdef __CINT__

#include <string>

#pragma link off all globals;
#pragma link off all classes;
#pragma link off all functions;

//#pragma link C++ namespace HDTV;
//#pragma link C++ namespace HDTV::Fit;
#pragma link C++ class HDTV::Fit::Integral+;
#pragma link C++ class HDTV::Fit::TH1Integral+;
#pragma link C++ class HDTV::Fit::BgIntegral+;
#pragma link C++ class HDTV::Fit::TH1BgsubIntegral+;
#pragma link C++ class HDTV::Fit::Background+;
#pragma link C++ class HDTV::Fit::ExpBg+;
#pragma link C++ class HDTV::Fit::PolyBg+;
#pragma link C++ class HDTV::Fit::InterpolationBg+;
#pragma link C++ class HDTV::Fit::Param+;
#pragma link C++ class HDTV::Fit::Option<bool>+;
#pragma link C++ class HDTV::Fit::Option<std::string>+;
#pragma link C++ class HDTV::Fit::Fitter+;
#pragma link C++ class HDTV::Fit::TheuerkaufPeak+;
#pragma link C++ class HDTV::Fit::TheuerkaufFitter+;
#pragma link C++ class HDTV::Fit::EEPeak+;
#pragma link C++ class HDTV::Fit::EEFitter+;
#pragma link C++ function HDTV::TH1IntegrateWithPartialBins;

#endif
