// clang-format off
#ifdef __CINT__

#pragma link off all globals;
#pragma link off all classes;
#pragma link off all functions;

#pragma link C++ namespace HDTV;
#pragma link C++ namespace HDTV::Display;
#pragma link C++ class HDTV::Display::View+;
#pragma link C++ class HDTV::Display::View1D+;
#pragma link C++ class HDTV::Display::View2D+;
#pragma link C++ class HDTV::Display::DisplaySpec;
#pragma link C++ class HDTV::Display::DisplayFunc;
#pragma link C++ class HDTV::Display::DisplayBlock;
#pragma link C++ class HDTV::Display::DisplayObj;
#pragma link C++ class HDTV::Display::Marker;
#pragma link C++ class HDTV::Display::XMarker;
#pragma link C++ class HDTV::Display::YMarker;
#pragma link C++ class HDTV::Display::MTViewer+;
#pragma link C++ class HDTV::Display::Viewer+;

#endif
